
/************************************************************************/
/* Copyright (c)  Marnix van den Bos.                        */
/*                                                                      */
/* <marnix@xvan.nl>                                                     */
/*                                                                      */
/* This file is part of XVAN, an Interactive Fiction authoring system.  */
/*                                                                      */
/* XVAN is free software: you can redistribute it and/or modify         */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, version 3.                             */
/*                                                                      */
/* XVAN is distributed in the hope that it will be useful,              */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* You should have received a copy of the GNU General Public License    */
/* along with XVAN.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                      */
/************************************************************************/


#include <stdio.h>
#include <stdlib.h>   /* malloc() */
#include <time.h>
#include <string.h>
#include <stdint.h>

#include "which_os.h"

#ifdef __windows_os
#include <conio.h>    /* getch()  */
#endif

#ifdef __osx_os
#include <sys/param.h>    /* realpath()            */
#include <limits.h>       /* PATH_MAX              */
#include <mach-o/dyld.h>  /* _NSGetExectablePath() */
#endif


#include "keyword.h"
#include "typedefs.h"
#include "init.h"

/***********/
/* Globals */
/***********/

FILE         *datafile       = NULL;  /* storyfile             */
FILE         *transcriptfile = NULL;  /* to store transcript   */
FILE         *testfile       = NULL;  /* to read testinput     */
/* deleted debugfile */

short        debug_info      = 0;
short        debug_level     = 0;

storyInfo    story_info;
int32_t      *stack;     /* Needed for executing triggers. */
int32_t      sp = 0;     /* stackpointer                   */

char         *outputbuffer = NULL;

choiceStruct choices[MAX_CHOICES];    /* for choice and hybrid mode */

/*************************/
/* function declarations */
/*************************/

int32_t CompilerVersionOK(void);
int32_t ExitProgram(void);
void    GetFileName(int, char**, char*, int*);
int32_t main(int, char**);

#ifdef __osx_os
char *GetFullPath(const char*);
#endif

/************************/
/* Function definitions */
/************************/

int32_t CompilerVersionOK(void)
{
  if (strcmp(story_info.compiler_version, "2.6") != 0) {  /* @!@ */
    PrintError(37, NULL, NULL);
    PrintError(38, NULL, story_info.compiler_version);
    return(ERROR);
  }

  /* correct compiler version */
  return(OK);
}


int32_t ExitProgram(void)
{
  /* we need to exit, but don't close the console window immediately */

  /* flush outputbuffer */
  Output(outputbuffer);

  /* close transcript file */
  if (transcript) {
    transcript = 0;
    fclose(transcriptfile);
  }

  PrintError(39, NULL, NULL);

#ifdef __windows_os
  getch();
#else
  getchar();
#endif

  /* Free all Mallocs */
  CleanUp();

  return(OK);
}


#ifdef __osx_os
char *GetFullPath(const char *filename)
{
  char      path_buf[PATH_MAX + 1];
  char      resolved_name[PATH_MAX + 1];
  char      *real_path;
  char      *return_path;
  uint32_t  buf_size = sizeof(path_buf);
  int       index    = 1;

  /* this functions returns the full path of the current */
  /* running application appended with var 'filename' at */
  /* the end. In case of an error it returns NULL.       */

  if ((return_path = (char *) malloc(MAX_FILENAME_LEN)) == NULL) {
    PrintError(15, NULL, "GetFullPath()");
    return(NULL);
  }

  /* get relative path */
  if (_NSGetExecutablePath(path_buf, &buf_size) != 0) {
    /* buffer too small */
    printf("File Path too long.");
    free(return_path);
    return(NULL);
  }

  /* convert to absolute path */
  if ( (real_path = realpath(path_buf, resolved_name)) == NULL) {
    printf("Could not determine path.\n");
    free(return_path);
    return(NULL);
  }

  /* strip the application name from the end of the path */
  index = strlen(real_path) - 1;
  while (real_path[index] != '/') {
    index--;
  }

  /* now check if there's enough room in return_path */
  if (strlen(real_path) + strlen(filename) >= MAX_FILENAME_LEN) {
    printf("File path too long.\n");
    free(return_path);
    return(NULL);
  }

  /* there's enough room, copy path and filename to return_path */
  strncpy(return_path, real_path, index+1);
  /* do not try to free() real_path */
  return_path[index+1] = '\0';
  strncat(return_path, filename, strlen(filename));

  return(return_path);  /* caller must free() return_path */
}
#endif


void GetFileName(int nr_args, char **args, char *filename, int *json)
{
  int i = 1;

#ifdef __osx_os
  char *path;
#endif

  filename[0]  = '\0';

  /* the command 'interpreter' itself is stored in argv[0] */

  while (i<nr_args) {
    if (strcmp(args[i], "-e") == 0) {
      /* we cannot set the json flag here, because */
      /* flags have not been initialized yet       */
      *json = 1;
      i++;
    }
    else {
      if (strcmp(args[i], "-i") == 0) {
        if (nr_args > i+1) {
          /* inputfile was specified */
          strncpy(filename, args[++i], MAX_ID_LENGTH);
          i++;
        }
      }
      else {
        PrintError(41, NULL, args[i]);
        i++;
      }
    }
  }

  if (filename[0] == '\0') {
    printf("\n\nInputfile: ");
    scanf("%19s", filename);

    filename[MAX_FILENAME_LEN] = '\0';

#ifdef __osx_os
    path = GetFullPath(filename);
    if (path != NULL) {
      strcpy(filename, path);
      free(path);
    }
    /* no need to free() path here */
#endif

    /* clear the input buffer, to prevent parsing an */
    /* empty string as the first user command        */
    /* always do a getchar() independent of OS */
    getchar();

    printf("\n");
  }
}

/************************/
/*         main()       */
/************************/

int32_t main(int argc, char **argv)
{
  char inputfile[MAX_ID_LENGTH+1];
  int json = 0;

  int i    = 0;

#ifdef __linux_os
  /* set background blue and text white for linux */
  printf("\x1b[44m");  /* linux */
  printf("\x1b[37m");
#endif

  /* init random number generator */
  srand((unsigned) time(NULL));

  /* malloc space for stack */
  if ((stack = (int32_t *) malloc(STACK_SIZE*sizeof(int32_t))) == NULL) {
    PrintError(15, NULL, "main()");
    ExitProgram();
    return(OK);
  }

  /* init the choice array                 */
  /* must do that here, because GetInput() */
  /* will use ResetString() onm it.        */
  for (i=0; i<MAX_CHOICES;i++) {
    choices[i].choice   = NULL;
    choices[i].response = NULL;
  }

  /* Get filename */
  GetFileName(argc, argv, inputfile, &json);

  /* Open the datafile */
  if ((datafile = fopen(inputfile, "rb")) == NULL) {
    PrintError(40, NULL, inputfile);
    ExitProgram();
    return(OK);
  }

  if (!ReadStoryInfo(&story_info)) {
    /* error was printed by ReadStoryInfo() */
    ExitProgram();
    return(OK);
  }

  /* First, check the compiler version in the story_info struct */
  if (!CompilerVersionOK()) {
    ExitProgram();
    return(OK);
  }

  /* Read the directories from the datafile */
  if (!InitDirs()) {
    ExitProgram();
    return(OK);
  }

  /* check for json mode */
  if (json == 1) {
    ProcLFlagVal(JSON_MODE, 1);
  }

  /* Read the max possible number of verb default codes.  */
  /* Can only do so AFTER the verb dir has been read.     */
  if (!InitVerbs()) {
    ExitProgram();
    return(OK);
  }

  /* Read the max possible number of locations. */
  /* Can only do so AFTER the location dir has been read. */
  if (!InitLocations()) {
    ExitProgram();
    return(OK);
  }

  /* Read the max possible number of objects. */
  /* Can only do so AFTER the object dir has been read. */
  if (!InitObjects()) {
    ExitProgram();
    return(OK);
  }

  /* Read attributes.                    */
  /* Can only do this AFTER InitDirs(), */
  /* InitLocations() and InitObjects(). */
  if (!InitAttributes()) {
    ExitProgram();
    return(OK);
  }

  /* initialize the undo stack  */
  if (!InitUndoStack()) {
    ExitProgram();
    return(OK);
  }

  /* OK, we're all set. Let's roll! */
  LetsPlay();
  ExitProgram();
  return(OK);
}
